"""Type definitions"""
from typing import List, Union, Callable, Awaitable, Any, Dict
import ssl as ssl_module

import aiohttp

from .constants import ConnectionType


#: Coroutine function
CoroFunction = Callable[..., Awaitable[Any]]
#: JSON object value
JsonObject = Dict[str, Any]
#: JSON serializer function
JsonDumper = Callable[[JsonObject], str]
#: JSON deserializer function
JsonLoader = Callable[[str], JsonObject]
#: Message payload (list of messages)
Payload = List[JsonObject]
#: Header values
Headers = Dict[str, str]
#: Connection type specification
ConnectionTypeSpec = Union[ConnectionType, List[ConnectionType]]
#: SSL validation mode
SSLValidationMode = Union[ssl_module.SSLContext, aiohttp.Fingerprint, bool]



"""Utility functions"""
import re
import asyncio
from functools import wraps
from http import HTTPStatus
from typing import Union, Optional, List, Any

from .constants import META_CHANNEL_PREFIX, SERVICE_CHANNEL_PREFIX
from .typing import CoroFunction, JsonObject


def defer(coro_func: CoroFunction, delay: Union[int, float, None] = None, *,
          loop: Optional[asyncio.AbstractEventLoop] = None) -> CoroFunction:
    """Returns a coroutine function that will defer the call to the given
    *coro_func* by *delay* seconds

    :param coro_func: A coroutine function
    :param delay: Delay in seconds
    :param loop: An event loop
    :return: Coroutine function wrapper
    """
    @wraps(coro_func)
    async def wrapper(*args: Any, **kwargs: Any) -> Any:  \
            # pylint: disable=missing-docstring
        if delay:
            await asyncio.sleep(delay)  # type: ignore
        return await coro_func(*args, **kwargs)

    return wrapper


def get_error_code(error_field: Union[str, None]) -> Optional[int]:
    """Get the error code part of the `error\
    <https://docs.cometd.org/current/reference/#_code_error_code>`_, message \
    field

    :param error_field: `Error\
    <https://docs.cometd.org/current/reference/#_code_error_code>`_, message \
    field
    :return: The error code as an int if 3 digits can be matched at the \
    beginning of the error field, for all other cases (``None`` or invalid \
    error field) return ``None``
    """
    result = None
    if error_field is not None:
        match = re.search(r"^\d{3}", error_field)
        if match:
            result = int(match[0])
    return result


def get_error_message(error_field: Union[str, None]) -> Optional[str]:
    """Get the description part of the `error\
    <https://docs.cometd.org/current/reference/#_code_error_code>`_, message \
    field

    :param error_field: `Error\
    <https://docs.cometd.org/current/reference/#_code_error_code>`_, message \
    field
    :return: The third part of the error field as a string if it can be \
    matched otherwise return ``None``
    """
    result = None
    if error_field is not None:
        match = re.search(r"(?<=:)[^:]*$", error_field)
        if match:
            result = match[0]
    return result


def get_error_args(error_field: Union[str, None]) -> Optional[List[str]]:
    """Get the arguments part of the `error\
    <https://docs.cometd.org/current/reference/#_code_error_code>`_, message \
    field

    :param error_field: `Error\
    <https://docs.cometd.org/current/reference/#_code_error_code>`_, message \
    field
    :return: The second part of the error field as a list of strings if it \
    can be matched otherwise return ``None``
    """
    result = None
    if error_field is not None:
        match = re.search(r"(?<=:).*(?=:)", error_field)
        if match:
            if match[0]:
                result = match[0].split(",")
            else:
                result = []
    return result


def is_matching_response(response_message: JsonObject,
                         message: Optional[JsonObject]) -> bool:
    """Check whether the *response_message* is a response for the
    given *message*.

    :param message: A sent message
    :param response_message: A response message
    :return: True if the *response_message* is a match for *message*
             otherwise False.
    """
    if message is None or response_message is None:
        return False
    # to consider a response message as a pair of the sent message
    # their channel should match, if they contain an id field it should
    # also match (according to the specs an id is always optional),
    # and the response message should contain the successful field
    return (message["channel"] == response_message["channel"] and
            message.get("id") == response_message.get("id") and
            "successful" in response_message)


def is_server_error_message(response_message: JsonObject) -> bool:
    """Check whether the *response_message* is a server side error message

    :param response_message: A response message
    :return: True if the *response_message* is a server side error message
             otherwise False.
    """
    return not response_message.get("successful", True)


def is_event_message(response_message: JsonObject) -> bool:
    """Check whether the *response_message* is an event message

    :param response_message: A response message
    :return: True if the *response_message* is an event message
             otherwise False.
    """
    channel = response_message["channel"]
    # every message is a response message if it's not on a meta channel
    # and if it's either not on a service channel, or it's on a service channel
    # but doesn't has an id (which means that it's not a response for an
    # outgoing message) and it has a data field
    return (not channel.startswith(META_CHANNEL_PREFIX) and
            (not channel.startswith(SERVICE_CHANNEL_PREFIX) or
             "id" not in response_message) and
            "data" in response_message)


def is_auth_error_message(response_message: JsonObject) -> bool:
    """Check whether the *response_message* is an authentication error
    message

    :param response_message: A response message
    :return: True if the *response_message* is an authentication error \
    message, otherwise False.
    """
    error_code = get_error_code(response_message.get("error"))
    # Strictly speaking, only UNAUTHORIZED should be considered as an auth
    # error, but some channels can also react with FORBIDDEN for auth
    # failures. This is certainly true for /meta/handshake, and since this
    # might happen for other channels as well, it's better to stay safe
    # and treat FORBIDDEN also as a potential auth error
    return error_code in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN)



"""CometD client for asyncio"""
import logging

from .client import Client  # noqa: F401
from .constants import ConnectionType  # noqa: F401
from .extensions import Extension, AuthExtension  # noqa: F401
from . import transports  # noqa: F401

# Create a default handler to avoid warnings in applications without logging
# configuration
logging.getLogger(__name__).addHandler(logging.NullHandler())



"""Exception types

Exception hierarchy::

    AiocometdException
        ClientError
            ClientInvalidOperation
        TransportError
            TransportInvalidOperation
            TransportTimeoutError
            TransportConnectionClosed
        ServerError
"""
from typing import Optional, List, cast

from .utils import JsonObject, get_error_code, get_error_message, get_error_args


class AiocometdException(Exception):
    """Base exception type.

    All exceptions of the package inherit from this class.
    """


class TransportError(AiocometdException):
    """Error during the transportation of messages"""


class TransportInvalidOperation(TransportError):
    """The requested operation can't be executed on the current state of the
    transport"""


class TransportTimeoutError(TransportError):
    """Transport timeout"""


class TransportConnectionClosed(TransportError):
    """The connection unexpectedly closed"""


class ServerError(AiocometdException):
    """CometD server side error"""
    # pylint: disable=useless-super-delegation
    def __init__(self, message: str, response: Optional[JsonObject]) \
            -> None:
        """If the *response* contains an error field it gets parsed
        according to the \
        `specs <https://docs.cometd.org/current/reference/#_code_error_code>`_

        :param message: Error description
        :param response: Server response message
        """
        super().__init__(message, response)

    # pylint: enable=useless-super-delegation

    @property
    def message(self) -> str:
        """Error description"""
        # pylint: disable=unsubscriptable-object
        return cast(str, self.args[0])
        # pylint: enable=unsubscriptable-object

    @property
    def response(self) -> Optional[JsonObject]:
        """Server response message"""
        return cast(Optional[JsonObject],
                    self.args[1])  # pylint: disable=unsubscriptable-object

    @property
    def error(self) -> Optional[str]:
        """Error field in the :obj:`response`"""
        if self.response is None:
            return None
        return self.response.get("error")

    @property
    def error_code(self) -> Optional[int]:
        """Error code part of the error code part of the `error\
        <https://docs.cometd.org/current/reference/#_code_error_code>`_, \
        message field"""
        return get_error_code(self.error)

    @property
    def error_message(self) -> Optional[str]:
        """Description part of the `error\
        <https://docs.cometd.org/current/reference/#_code_error_code>`_, \
        message field"""
        return get_error_message(self.error)

    @property
    def error_args(self) -> Optional[List[str]]:
        """Arguments part of the `error\
        <https://docs.cometd.org/current/reference/#_code_error_code>`_, \
        message field"""
        return get_error_args(self.error)


class ClientError(AiocometdException):
    """ComtedD client side error"""


class ClientInvalidOperation(ClientError):
    """The requested operation can't be executed on the current state of the
    client"""



"""Extension classes"""
from abc import ABC, abstractmethod
from typing import Optional

from .typing import Payload, Headers


class Extension(ABC):
    """Defines operations supported by extensions"""
    @abstractmethod
    async def outgoing(self, payload: Payload, headers: Headers) -> None:
        """Process outgoing *payload* and *headers*

        Called just before a payload is sent

        :param payload: List of outgoing messages
        :param headers: Headers to send
        """

    @abstractmethod
    async def incoming(self, payload: Payload,
                       headers: Optional[Headers] = None) -> None:
        """Process incoming *payload* and *headers*

        Called just after a payload is received

        :param payload: List of incoming messages
        :param headers: Headers to send
        """


class AuthExtension(Extension):  # pylint: disable=abstract-method
    """Extension with support for authentication"""
    async def authenticate(self) -> None:
        """Called after a failed authentication attempt

        For authentication schemes where the credentials are static it doesn't
        makes much sense to reimplement this function. However for schemes
        where the credentials can expire (like OAuth, JWT...) this method can
        be reimplemented to update those credentials
        """



"""Transport abstract base class definition"""
from abc import ABC, abstractmethod
from typing import Set, Optional, List

from ..constants import ConnectionType, TransportState
from ..typing import JsonObject


class Transport(ABC):
    """Defines the operations that all transport classes should support"""
    @property
    @abstractmethod
    def connection_type(self) -> ConnectionType:
        """The transport's connection type"""

    @property
    @abstractmethod
    def endpoint(self) -> str:
        """CometD service url"""

    @property
    @abstractmethod
    def client_id(self) -> Optional[str]:
        """Clinet id value assigned by the server"""

    @property
    @abstractmethod
    def state(self) -> TransportState:
        """Current state of the transport"""

    @property
    @abstractmethod
    def subscriptions(self) -> Set[str]:
        """Set of subscribed channels"""

    @property
    @abstractmethod
    def last_connect_result(self) -> Optional[JsonObject]:
        """Result of the last connect request"""

    @property
    @abstractmethod
    def reconnect_advice(self) -> JsonObject:
        """Reconnection advice parameters returned by the server"""

    @abstractmethod
    async def handshake(self, connection_types: List[ConnectionType]) \
            -> JsonObject:
        """Executes the handshake operation

        :param connection_types: list of connection types
        :return: Handshake response
        :raises TransportError: When the network request fails.
        """

    @abstractmethod
    async def connect(self) -> JsonObject:
        """Connect to the server

        The transport will try to start and maintain a continuous connection
        with the server, but it'll return with the response of the first
        successful connection as soon as possible.

        :return: The response of the first successful connection.
        :raise TransportInvalidOperation: If the transport doesn't has a \
        client id yet, or if it's not in a :obj:`~TransportState.DISCONNECTED`\
        :obj:`state`.
        :raises TransportError: When the network request fails.
        """

    @abstractmethod
    async def disconnect(self) -> None:
        """Disconnect from server

        The disconnect message is only sent to the server if the transport is
        actually connected.
        """

    @abstractmethod
    async def close(self) -> None:
        """Close transport and release resources"""

    @abstractmethod
    async def subscribe(self, channel: str) -> JsonObject:
        """Subscribe to *channel*

        :param channel: Name of the channel
        :return: Subscribe response
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """

    @abstractmethod
    async def unsubscribe(self, channel: str) -> JsonObject:
        """Unsubscribe from *channel*

        :param channel: Name of the channel
        :return: Unsubscribe response
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """

    @abstractmethod
    async def publish(self, channel: str, data: JsonObject) -> JsonObject:
        """Publish *data* to the given *channel*

        :param channel: Name of the channel
        :param data: Data to send to the server
        :return: Publish response
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """

    @abstractmethod
    async def wait_for_state(self, state: TransportState) -> None:
        """Waits for and returns when the transport enters the given *state*

        :param state: A state value
        """



"""Functions for transport class registration and instantiation"""
from typing import Type, Callable, Any

from ..exceptions import TransportInvalidOperation
from ..constants import ConnectionType
from .abc import Transport


TRANSPORT_CLASSES = {}


def register_transport(conn_type: ConnectionType) \
        -> Callable[[Type[Transport]], Type[Transport]]:
    """Class decorator for registering transport classes

    The class' connection_type property will be also defined to return the
    given *connection_type*
    :param conn_type: A connection type
    :return: The updated class
    """
    # pylint: disable=unused-argument, missing-docstring
    def decorator(cls: Type[Transport]) -> Type[Transport]:
        TRANSPORT_CLASSES[conn_type] = cls

        @property  # type: ignore
        def connection_type(self: Transport) -> ConnectionType:
            return conn_type

        cls.connection_type = connection_type  # type: ignore
        return cls
    return decorator


def create_transport(connection_type: ConnectionType, *args: Any,
                     **kwargs: Any) -> Transport:
    """Create a transport object that can be used for the given
    *connection_type*

    :param connection_type: A connection type
    :param kwargs: Keyword arguments to pass to the transport
    :return: A transport object
    """
    if connection_type not in TRANSPORT_CLASSES:
        raise TransportInvalidOperation("There is no transport for connection "
                                        "type {!r}".format(connection_type))

    return TRANSPORT_CLASSES[connection_type](*args, **kwargs)  # type: ignore



"""Websocket transport class definition"""
import asyncio
import logging
from contextlib import suppress
from typing import Callable, Optional, AsyncContextManager, Any, Awaitable, \
    cast, Dict

import aiohttp
import aiohttp.client_ws

from ..constants import ConnectionType
from ..exceptions import TransportError, TransportConnectionClosed
from ..typing import JsonObject
from .registry import register_transport
from .base import TransportBase, Payload, Headers


LOGGER = logging.getLogger(__name__)
#: Asynchronous factory function of ClientSessions
AsyncSessionFactory = Callable[[], Awaitable[aiohttp.ClientSession]]
#: Web socket type
WebSocket = aiohttp.client_ws.ClientWebSocketResponse
#: Context manager type managing a WebSocket
WebSocketContextManager = AsyncContextManager[WebSocket]


class WebSocketFactory:  # pylint: disable=too-few-public-methods
    """Helper class to create asynchronous callable objects, that return
    WebSocket objects

    This class allows the usage of WebSocket objects without context blocks
    """
    def __init__(self, http_session: aiohttp.ClientSession):
        """
        :param http_session: HTTP session
        """
        self._http_session = http_session
        self._context: Optional[WebSocketContextManager] = None
        self._socket: Optional[WebSocket] = None

    async def close(self) -> None:
        """Close the WebSocket"""
        with suppress(Exception):
            await self._exit()

    async def __call__(self, *args: Any, **kwargs: Any) -> WebSocket:
        """Create a new WebSocket object or return a previously created one
        if it's not closed

        :param args: positional arguments for the ws_connect function
        :param kwargs: keyword arguments for the ws_connect function
        :return: Websocket object
        """
        # if a the factory object already exists and if it's in closed state
        # exit the context manager and clear the references
        if self._socket is not None and self._socket.closed:
            await self._exit()

        # if there is no factory object, then create it and enter the \
        # context manager to initialize it
        if self._socket is None:
            self._socket = await self._enter(*args, **kwargs)

        return self._socket

    async def _enter(self, *args: Any, **kwargs: Any) -> WebSocket:
        """Enter WebSocket context

        :param args: positional arguments for the ws_connect function
        :param kwargs: keyword arguments for the ws_connect function
        :return: Websocket object
        """
        self._context = self._http_session.ws_connect(*args, **kwargs)
        return await self._context.__aenter__()

    async def _exit(self) -> None:
        """Exit WebSocket context"""
        if self._context:
            await self._context.__aexit__(None, None, None)
            self._socket = self._context = None


@register_transport(ConnectionType.WEBSOCKET)
class WebSocketTransport(TransportBase):
    """WebSocket type transport"""

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)
        #: factory for creating websockets
        self._socket_factory = WebSocketFactory(self._http_session)
        #: pending message exchanges between the client and server,
        #: the request message's id is used as a key
        self._pending_exhanges: Dict[int, "asyncio.Future[JsonObject]"] \
            = dict()
        #: task for receiving incoming messages
        self._receive_task: Optional["asyncio.Task[None]"] = None

    async def _reset_socket(self) -> None:
        """Close the socket factory and recreate it"""
        await self._socket_factory.close()
        self._socket_factory = WebSocketFactory(self._http_session)

    async def _get_socket(self, headers: Headers) -> WebSocket:
        """Factory function for creating a websocket object

        :param headers: Headers to send
        :return: Websocket object
        """
        return await self._socket_factory(
            self.endpoint,
            ssl=self.ssl,
            headers=headers,
            receive_timeout=self.request_timeout,
            autoping=True)

    def _create_exhange_future(self, payload: Payload) \
            -> "asyncio.Future[JsonObject]":
        """Create a future which represents an exchange of messages between
        the server and client

        The created future will be associated with the id of the first message
        in the payload.
        :param payload: The payload sent by the client
        :return: A future which will yield the server's response message to the
        outgoing *payload*
        """
        future: "asyncio.Future[JsonObject]" = asyncio.Future(loop=self._loop)
        self._pending_exhanges[payload[0]["id"]] = future
        return future

    def _set_exchange_results(self, response_payload: Payload) -> None:
        """Set the result of all the pending message exchange futures for which
        we can find a response in the payload

        :param response_payload: Response payload
        """
        # iterate over all incoming messages
        for response_message in response_payload:
            # if the incoming message has an id (otherwise it's not a response)
            if "id" in response_message:
                message_id = response_message["id"]
                # if the message id is associated with any pending exchange
                if message_id in self._pending_exhanges:
                    # remove the exchange from the pending exchanges
                    exchange = self._pending_exhanges.pop(message_id)
                    # if the future is not completed yet then set its result
                    if not exchange.done():
                        exchange.set_result(response_message)

    def _set_exchange_errors(self, error: Exception) -> None:
        """Set the *error* as the exception for all pending exchanges

        :param error: An exception
        """
        # set the exception for all the exchanges
        for exchange in self._pending_exhanges.values():
            if not exchange.done():
                exchange.set_exception(error)
        # clear the pending exchanges
        self._pending_exhanges.clear()

    async def _send_final_payload(self, payload: Payload, *,
                                  headers: Headers) -> JsonObject:
        try:
            try:
                # try to send the payload on the socket which might have
                # been closed since the last time it was used
                socket = await self._get_socket(headers)
                return await self._send_socket_payload(socket, payload)
            except asyncio.TimeoutError:
                # reset the socket factory since after a timeout error
                # it becomes invalid
                await self._reset_socket()
                raise
            except TransportConnectionClosed:
                # if the socket was indeed closed, try to reopen the socket
                # and send the payload, since the connection could've
                # normalised since the last network problem
                socket = await self._get_socket(headers)
                return await self._send_socket_payload(socket, payload)
        except aiohttp.client_exceptions.ClientError as error:
            LOGGER.warning("Failed to send payload, %s", error)
            raise TransportError(str(error)) from error

    async def _send_socket_payload(self, socket: WebSocket,
                                   payload: Payload) -> JsonObject:
        """Send *payload* to the server on the given *socket*

        :param socket: WebSocket object
        :param payload: A message or a list of messages
        :return: Response payload
        :raises TransportError: When the request fails.
        :raises TransportConnectionClosed: When the *socket* receives a CLOSE \
        message instead of the expected response
        """
        # create a future for the exchange of messages
        future = self._create_exhange_future(payload)
        try:
            # send the outgoing payload
            await socket.send_json(payload, dumps=self._json_dumps)
        except Exception as error:
            # set the error as the result for all pending exchanges
            self._set_exchange_errors(error)
            raise

        # make sure the receive task is running
        self._start_receive_task(socket)
        # await and return the response of the server
        return await future

    def _start_receive_task(self, socket: WebSocket) -> None:
        """Start the task which receives messages from the *socket* if it's
        not already running

        :param socket: A Websocket object
        """
        # if the receive task is not running then start it
        if self._receive_task is None:
            self._receive_task = self._loop.create_task(self._receive(socket))
            self._receive_task.add_done_callback(self._receive_done)

    async def _receive(self, socket: WebSocket) -> None:
        """Consume the incomming messages on the given *socket*

        :param socket: A Websocket object
        """
        # receive responses from the server and consume them
        try:
            while True:
                response = await socket.receive()
                if response.type == aiohttp.WSMsgType.CLOSE:
                    raise TransportConnectionClosed("Received CLOSE message "
                                                    "on the factory.")
                # parse the response payload
                try:
                    response_payload \
                        = cast(Payload, response.json(loads=self._json_loads))
                except TypeError:
                    raise TransportError("Received invalid response from the "
                                         "server.")

                # consume all event messages in the payload
                await self._consume_payload(response_payload)

                # set results of matching exchanges
                self._set_exchange_results(response_payload)
        except Exception as error:
            # set the error as the result for all pending exchanges
            self._set_exchange_errors(error)
            raise

    def _receive_done(self, future: "asyncio.Task[None]") -> None:
        """Consume the results of the *future*

        :param future: A :obj:`_receive` future
        """
        # extract the result of the future
        try:
            result = future.result()
        except Exception as error:  # pylint: disable=broad-except
            result = error
        # clear the receive task
        self._receive_task = None
        LOGGER.debug("Recevie task finished with: %r", result)

    async def close(self) -> None:
        # cancel the receive task if it exists and wait for its completeion
        if self._receive_task is not None and not self._receive_task.done():
            self._receive_task.cancel()
            await asyncio.wait([self._receive_task])
        await self._socket_factory.close()
        await super().close()



"""Client class implementation"""
import asyncio
import reprlib
import logging
from collections import abc
from contextlib import suppress
import json
from typing import Optional, List, Union, Set, AsyncIterator, Type, Any
from types import TracebackType

import aiohttp

from .transports import create_transport
from .transports.abc import Transport
from .constants import DEFAULT_CONNECTION_TYPE, ConnectionType, \
    MetaChannel, SERVICE_CHANNEL_PREFIX, TransportState
from .exceptions import ServerError, ClientInvalidOperation, \
    TransportTimeoutError, ClientError
from .utils import is_server_error_message
from .extensions import Extension, AuthExtension
from .typing import ConnectionTypeSpec, SSLValidationMode, \
    JsonObject, JsonDumper, JsonLoader


LOGGER = logging.getLogger(__name__)


class Client:  # pylint: disable=too-many-instance-attributes
    """CometD client"""
    #: Predefined server error messages by channel name
    _SERVER_ERROR_MESSAGES = {
        MetaChannel.HANDSHAKE: "Handshake request failed.",
        MetaChannel.CONNECT: "Connect request failed.",
        MetaChannel.DISCONNECT: "Disconnect request failed.",
        MetaChannel.SUBSCRIBE: "Subscribe request failed.",
        MetaChannel.UNSUBSCRIBE: "Unsubscribe request failed."
    }
    #: Defualt connection types list
    _DEFAULT_CONNECTION_TYPES = [ConnectionType.WEBSOCKET,
                                 ConnectionType.LONG_POLLING]
    #: Timeout to give to HTTP session to close itself
    _HTTP_SESSION_CLOSE_TIMEOUT = 0.250

    def __init__(self, url: str,
                 connection_types: Optional[ConnectionTypeSpec] = None, *,
                 connection_timeout: Union[int, float] = 10.0,
                 ssl: Optional[SSLValidationMode] = None,
                 max_pending_count: int = 100,
                 extensions: Optional[List[Extension]] = None,
                 auth: Optional[AuthExtension] = None,
                 json_dumps: JsonDumper = json.dumps,
                 json_loads: JsonLoader = json.loads,
                 loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        """
        :param url: CometD service url
        :param connection_types: List of connection types in order of \
        preference, or a single connection type name. If ``None``, \
        [:obj:`~ConnectionType.WEBSOCKET`, \
        :obj:`~ConnectionType.LONG_POLLING`] will be used as a default value.
        :param connection_timeout: The maximum amount of time to wait for the \
        transport to re-establish a connection with the server when the \
        connection fails.
        :param ssl: SSL validation mode. None for default SSL check \
        (:func:`ssl.create_default_context` is used), False for skip SSL \
        certificate validation, \
        `aiohttp.Fingerprint <https://aiohttp.readthedocs.io/en/stable/\
        client_reference.html#aiohttp.Fingerprint>`_ for fingerprint \
        validation, :obj:`ssl.SSLContext` for custom SSL certificate \
        validation.
        :param max_pending_count: The maximum number of messages to \
        prefetch from the server. If the number of prefetched messages reach \
        this size then the connection will be suspended, until messages are \
        consumed. \
        If it is less than or equal to zero, the count is infinite.
        :param extensions: List of protocol extension objects
        :param auth: An auth extension
        :param json_dumps: Function for JSON serialization, the default is \
        :func:`json.dumps`
        :param json_loads: Function for JSON deserialization, the default is \
        :func:`json.loads`
        :param loop: Event :obj:`loop <asyncio.BaseEventLoop>` used to
                     schedule tasks. If *loop* is ``None`` then
                     :func:`asyncio.get_event_loop` is used to get the default
                     event loop.
        """
        #: CometD service url
        self.url = url
        #: List of connection types to use in order of preference
        if isinstance(connection_types, ConnectionType):
            self._connection_types = [connection_types]
        elif isinstance(connection_types, abc.Iterable):
            self._connection_types = list(connection_types)
        else:
            self._connection_types = self._DEFAULT_CONNECTION_TYPES
        self._loop = loop or asyncio.get_event_loop()
        #: queue for consuming incoming event messages
        self._incoming_queue: "Optional[asyncio.Queue[JsonObject]]" = None
        #: transport object
        self._transport: Optional[Transport] = None
        #: marks whether the client is open or closed
        self._closed = True
        #: The maximum amount of time to wait for the transport to re-establish
        #: a connection with the server when the connection fails
        self.connection_timeout = connection_timeout
        #: SSL validation mode
        self.ssl = ssl
        #: the maximum number of messages to prefetch from the server
        self._max_pending_count = max_pending_count
        #: List of protocol extension objects
        self.extensions = extensions
        #: An auth extension
        self.auth = auth
        #: Function for JSON serialization
        self._json_dumps = json_dumps
        #: Function for JSON deserialization
        self._json_loads = json_loads
        #: http session
        self._http_session: Optional[aiohttp.ClientSession] = None

    def __repr__(self) -> str:
        """Formal string representation"""
        cls_name = type(self).__name__
        fmt_spec = "{}({}, {}, connection_timeout={}, ssl={}, " \
                   "max_pending_count={}, extensions={}, auth={}, loop={})"
        return fmt_spec.format(cls_name,
                               reprlib.repr(self.url),
                               reprlib.repr(self._connection_types),
                               reprlib.repr(self.connection_timeout),
                               reprlib.repr(self.ssl),
                               reprlib.repr(self._max_pending_count),
                               reprlib.repr(self.extensions),
                               reprlib.repr(self.auth),
                               reprlib.repr(self._loop))

    @property
    def closed(self) -> bool:
        """Marks whether the client is open or closed"""
        return self._closed

    @property
    def subscriptions(self) -> Set[str]:
        """Set of subscribed channels"""
        if self._transport:
            return self._transport.subscriptions
        return set()

    @property
    def connection_type(self) -> Optional[ConnectionType]:
        """The current connection type in use if the client is open,
        otherwise ``None``"""
        if self._transport is not None:
            return self._transport.connection_type
        return None

    @property
    def pending_count(self) -> int:
        """The number of pending incoming messages

        Once :obj:`open` is called the client starts listening for messages
        from the server. The incoming messages are retrieved and stored in an
        internal queue until they get consumed by calling :obj:`receive`.
        """
        if self._incoming_queue is None:
            return 0
        return self._incoming_queue.qsize()

    @property
    def has_pending_messages(self) -> bool:
        """Marks whether the client has any pending incoming messages"""
        return self.pending_count > 0

    async def _get_http_session(self) -> aiohttp.ClientSession:
        """Factory method for getting the current HTTP session

        :return: The current session if it's not None, otherwise it creates a
                 new session.
        """
        # it would be nicer to create the session when the class gets
        # initialized, but this seems to be the right way to do it since
        # aiohttp produces log messages with warnings that a session should be
        # created in a coroutine
        if self._http_session is None:
            self._http_session = aiohttp.ClientSession(
                json_serialize=self._json_dumps
            )
        return self._http_session

    async def _close_http_session(self) -> None:
        """Close the http session if it's not already closed"""
        # graceful shutdown recommended by the documentation
        # https://aiohttp.readthedocs.io/en/stable/client_advanced.html\
        # #graceful-shutdown
        if self._http_session is not None and not self._http_session.closed:
            await self._http_session.close()
            await asyncio.sleep(self._HTTP_SESSION_CLOSE_TIMEOUT)

    def _pick_connection_type(self, connection_types: List[str]) \
            -> Optional[ConnectionType]:
        """Pick a connection type based on the  *connection_types*
        supported by the server and on the user's preferences

        :param connection_types: Connection types \
        supported by the server
        :return: The connection type with the highest precedence \
        which is supported by the server
        """
        server_connection_types = []
        for type_string in connection_types:
            with suppress(ValueError):
                server_connection_types.append(ConnectionType(type_string))

        intersection = (set(server_connection_types) &
                        set(self._connection_types))
        if not intersection:
            return None

        result = min(intersection, key=self._connection_types.index)
        return result

    async def _negotiate_transport(self) -> Transport:
        """Negotiate the transport type to use with the server and create the
        transport object

        :return: Transport object
        :raise ClientError: If none of the connection types offered by the \
        server are supported
        """
        self._incoming_queue = asyncio.Queue(maxsize=self._max_pending_count)
        http_session = await self._get_http_session()
        transport = create_transport(DEFAULT_CONNECTION_TYPE,
                                     url=self.url,
                                     incoming_queue=self._incoming_queue,
                                     ssl=self.ssl,
                                     extensions=self.extensions,
                                     auth=self.auth,
                                     json_dumps=self._json_dumps,
                                     json_loads=self._json_loads,
                                     http_session=http_session,
                                     loop=self._loop)

        try:
            response = await transport.handshake(self._connection_types)
            self._verify_response(response)

            LOGGER.info("Connection types supported by the server: %r",
                        response["supportedConnectionTypes"])
            connection_type = self._pick_connection_type(
                response["supportedConnectionTypes"]
            )
            if not connection_type:
                raise ClientError("None of the connection types offered by "
                                  "the server are supported.")

            if transport.connection_type != connection_type:
                # extract and reuse the client_id from the initial transport
                client_id = transport.client_id
                # extract and reuse the reconnect_advice from the initial
                # transport
                advice = transport.reconnect_advice
                # close the initial transport
                await transport.close()
                # create the negotiated transport
                transport = create_transport(
                    connection_type,
                    url=self.url,
                    incoming_queue=self._incoming_queue,
                    client_id=client_id,
                    ssl=self.ssl,
                    extensions=self.extensions,
                    auth=self.auth,
                    json_dumps=self._json_dumps,
                    json_loads=self._json_loads,
                    reconnect_advice=advice,
                    http_session=http_session,
                    loop=self._loop)
            return transport
        except Exception:
            await transport.close()
            await self._close_http_session()
            raise

    async def open(self) -> None:
        """Establish a connection with the CometD server

        This method works mostly the same way as the `handshake` method of
        CometD clients in the reference implementations.

        :raise ClientError: If none of the connection types offered by the \
        server are supported
        :raise ClientInvalidOperation:  If the client is already open, or in \
        other words if it isn't :obj:`closed`
        :raise TransportError: If a network or transport related error occurs
        :raise ServerError: If the handshake or the first connect request \
        gets rejected by the server.
        """
        if not self.closed:
            raise ClientInvalidOperation("Client is already open.")

        LOGGER.info("Opening client with connection types %r ...",
                    [t.value for t in self._connection_types])
        self._transport = await self._negotiate_transport()

        response = await self._transport.connect()
        self._verify_response(response)
        self._closed = False

        assert self.connection_type is not None
        LOGGER.info("Client opened with connection_type %r",
                    self.connection_type.value)

    async def close(self) -> None:
        """Disconnect from the CometD server"""
        if not self.closed:
            if self.pending_count == 0:
                LOGGER.info("Closing client...")
            else:
                LOGGER.warning(
                    "Closing client while %s messages are still pending...",
                    self.pending_count)
            try:
                if self._transport:
                    await self._transport.disconnect()
                    await self._transport.close()
                await self._close_http_session()
            finally:
                self._closed = True
                LOGGER.info("Client closed.")

    async def subscribe(self, channel: str) -> None:
        """Subscribe to *channel*

        :param channel: Name of the channel
        :raise ClientInvalidOperation: If the client is :obj:`closed`
        :raise TransportError: If a network or transport related error occurs
        :raise ServerError: If the subscribe request gets rejected by the \
        server
        """
        if self.closed:
            raise ClientInvalidOperation("Can't send subscribe request while, "
                                         "the client is closed.")
        await self._check_server_disconnected()

        assert self._transport is not None
        response = await self._transport.subscribe(channel)
        self._verify_response(response)
        LOGGER.info("Subscribed to channel %s", channel)

    async def unsubscribe(self, channel: str) -> None:
        """Unsubscribe from *channel*

        :param channel: Name of the channel
        :raise ClientInvalidOperation: If the client is :obj:`closed`
        :raise TransportError: If a network or transport related error occurs
        :raise ServerError: If the unsubscribe request gets rejected by the \
        server
        """
        if self.closed:
            raise ClientInvalidOperation("Can't send unsubscribe request "
                                         "while, the client is closed.")
        await self._check_server_disconnected()

        assert self._transport is not None
        response = await self._transport.unsubscribe(channel)
        self._verify_response(response)
        LOGGER.info("Unsubscribed from channel %s", channel)

    async def publish(self, channel: str, data: JsonObject) -> JsonObject:
        """Publish *data* to the given *channel*

        :param channel: Name of the channel
        :param data: Data to send to the server
        :return: Publish response
        :raise ClientInvalidOperation: If the client is :obj:`closed`
        :raise TransportError: If a network or transport related error occurs
        :raise ServerError: If the publish request gets rejected by the server
        """
        if self.closed:
            raise ClientInvalidOperation("Can't publish data while, "
                                         "the client is closed.")
        await self._check_server_disconnected()

        assert self._transport is not None
        response = await self._transport.publish(channel, data)
        self._verify_response(response)
        return response

    def _verify_response(self, response: JsonObject) -> None:
        """Check the ``successful`` status of the *response* and raise \
        the appropriate :obj:`~aiocometd_ten.exceptions.ServerError` if it's False

        If the *response* has no ``successful`` field, it's considered to be
        successful.

        :param response: Response message
        :raise ServerError: If the *response* is not ``successful``
        """
        if is_server_error_message(response):
            self._raise_server_error(response)

    def _raise_server_error(self, response: JsonObject) -> None:
        """Raise the appropriate :obj:`~aiocometd_ten.exceptions.ServerError` for \
        the failed *response*

        :param response: Response message
        :raise ServerError: If the *response* is not ``successful``
        """
        channel = response["channel"]
        message = type(self)._SERVER_ERROR_MESSAGES.get(channel)
        if not message:
            if channel.startswith(SERVICE_CHANNEL_PREFIX):
                message = "Service request failed."
            else:
                message = "Publish request failed."
        raise ServerError(message, response)

    async def receive(self) -> JsonObject:
        """Wait for incoming messages from the server

        :return: Incoming message
        :raise ClientInvalidOperation: If the client is closed, and has no \
        more pending incoming messages
        :raise ServerError: If the client receives a confirmation message \
         which is not ``successful``
        :raise TransportTimeoutError: If the transport can't re-establish \
        connection with the server in :obj:`connection_timeout` time.
        """
        if not self.closed or self.has_pending_messages:
            response = await self._get_message(self.connection_timeout)
            self._verify_response(response)
            return response

        raise ClientInvalidOperation("The client is closed and there are "
                                     "no pending messages.")

    async def __aiter__(self) -> AsyncIterator[JsonObject]:
        """Asynchronous iterator

        :raise ServerError: If the client receives a confirmation message \
         which is not ``successful``
        :raise TransportTimeoutError: If the transport can't re-establish \
        connection with the server in :obj:`connection_timeout` time.
        """
        while True:
            try:
                yield await self.receive()
            except ClientInvalidOperation:
                break

    async def __aenter__(self) -> "Client":
        """Enter the runtime context and call :obj:`open`

        :raise ClientInvalidOperation:  If the client is already open, or in \
        other words if it isn't :obj:`closed`
        :raise TransportError: If a network or transport related error occurs
        :raise ServerError: If the handshake or the first connect request \
        gets rejected by the server.
        :return: The client object itself
        :rtype: Client
        """
        try:
            await self.open()
        except Exception:
            await self.close()
            raise
        return self

    async def __aexit__(self, exc_type: Type[BaseException],
                        exc_val: BaseException,
                        exc_tb: TracebackType) -> None:
        """Exit the runtime context and call :obj:`open`"""
        await self.close()

    async def _get_message(self, connection_timeout: Union[int, float]) \
            -> JsonObject:
        """Get the next incoming message

        :param connection_timeout: The maximum amount of time to wait for the \
        transport to re-establish a connection with the server when the \
        connection fails.
        :return: Incoming message
        :raise TransportTimeoutError: If the transport can't re-establish \
        connection with the server in :obj:`connection_timeout` time.
        :raise ServerError: If the connection gets closed by the server.
        """
        tasks: List[asyncio.Future[Any]] = []
        # task waiting on connection timeout
        if connection_timeout:
            timeout_task = asyncio.ensure_future(
                self._wait_connection_timeout(connection_timeout),
                loop=self._loop
            )
            tasks.append(timeout_task)

        assert self._incoming_queue is not None
        # task waiting on incoming messages
        get_task = asyncio.ensure_future(self._incoming_queue.get(),
                                         loop=self._loop)
        tasks.append(get_task)

        assert self._transport is not None
        # task waiting on server side disconnect
        server_disconnected_task = asyncio.ensure_future(
            self._transport.wait_for_state(
                TransportState.SERVER_DISCONNECTED),
            loop=self._loop
        )
        tasks.append(server_disconnected_task)

        try:
            done, pending = await asyncio.wait(
                tasks,
                return_when=asyncio.FIRST_COMPLETED)

            # cancel all pending tasks
            for task in pending:
                task.cancel()

            # handle the completed task
            if get_task in done:
                return get_task.result()

            if server_disconnected_task in done:
                await self.close()
                raise ServerError("Connection closed by the server",
                                  self._transport.last_connect_result)
            raise TransportTimeoutError("Lost connection with the server.")
        except asyncio.CancelledError:
            # cancel all tasks
            for task in tasks:
                task.cancel()
            raise

    async def _wait_connection_timeout(self, timeout: Union[int, float]) \
            -> None:
        """Wait for and return when the transport can't re-establish \
        connection with the server in *timeout* time

        :param timeout: The maximum amount of time to wait for the \
        transport to re-establish a connection with the server when the \
        connection fails.
        """
        assert self._transport is not None
        while True:
            await self._transport.wait_for_state(TransportState.CONNECTING)
            try:
                await asyncio.wait_for(
                    self._transport.wait_for_state(TransportState.CONNECTED),
                    timeout
                )
            except asyncio.TimeoutError:
                break

    async def _check_server_disconnected(self) -> None:
        """Checks whether the current transport'state is
        :obj:`TransportState.SERVER_DISCONNECTED` and if it is then closes the
        client and raises an error

        :raise ServerError: If the current transport's state is \
        :obj:`TransportState.SERVER_DISCONNECTED`
        """
        if (self._transport and
                self._transport.state == TransportState.SERVER_DISCONNECTED):
            await self.close()
            raise ServerError("Connection closed by the server",
                              self._transport.last_connect_result)



"""Base transtport abstract class definition"""
import asyncio
import logging
from abc import abstractmethod
from contextlib import suppress
import json
from typing import Union, Optional, List, Set, Awaitable, Any

import aiohttp

from ..constants import ConnectionType, MetaChannel, TransportState, \
    HANDSHAKE_MESSAGE, CONNECT_MESSAGE, DISCONNECT_MESSAGE, \
    SUBSCRIBE_MESSAGE, UNSUBSCRIBE_MESSAGE, PUBLISH_MESSAGE
from ..utils import defer, is_matching_response, \
    is_auth_error_message, is_event_message
from ..exceptions import TransportInvalidOperation, TransportError
from ..typing import SSLValidationMode, JsonObject, JsonLoader, \
    JsonDumper, Headers, Payload
from ..extensions import Extension, AuthExtension
from .abc import Transport


LOGGER = logging.getLogger(__name__)


class TransportBase(Transport):  # pylint: disable=too-many-instance-attributes
    """Base transport implementation

    This class contains most of the transport operations implemented, it can
    be used as a base class for various concrete transport implementations.
    When subclassing, at a minimum the :meth:`_send_final_payload` and
    :obj:`~Transport.connection_type` methods should be reimplemented.
    """
    #: The increase factor to use for request timeout
    REQUEST_TIMEOUT_INCREASE_FACTOR = 1.2

    def __init__(self, *, url: str,
                 incoming_queue: "asyncio.Queue[JsonObject]",
                 http_session: aiohttp.ClientSession,
                 client_id: Optional[str] = None,
                 reconnection_timeout: Union[int, float] = 1,
                 ssl: Optional[SSLValidationMode] = None,
                 extensions: Optional[List[Extension]] = None,
                 auth: Optional[AuthExtension] = None,
                 json_dumps: JsonDumper = json.dumps,
                 json_loads: JsonLoader = json.loads,
                 reconnect_advice: Optional[JsonObject] = None,
                 loop: Optional[asyncio.AbstractEventLoop] = None) -> None:
        """
        :param url: CometD service url
        :param incoming_queue: Queue for consuming incoming event
                                             messages
        :param client_id: Clinet id value assigned by the server
        :param reconnection_timeout: The time to wait before trying to \
        reconnect to the server after a network failure
        :param ssl: SSL validation mode. None for default SSL check \
        (:func:`ssl.create_default_context` is used), False for skip SSL \
        certificate validation, \
        `aiohttp.Fingerprint <https://aiohttp.readthedocs.io/en/stable/\
        client_reference.html#aiohttp.Fingerprint>`_ for fingerprint \
        validation, :obj:`ssl.SSLContext` for custom SSL certificate \
        validation.
        :param extensions: List of protocol extension objects
        :param auth: An auth extension
        :param json_dumps: Function for JSON serialization, the default is \
        :func:`json.dumps`
        :param json_loads: Function for JSON deserialization, the default is \
        :func:`json.loads`
        :param reconnect_advice: Initial reconnect advice
        :param http_session: HTTP client session
        :param loop: Event :obj:`loop <asyncio.BaseEventLoop>` used to
                     schedule tasks. If *loop* is ``None`` then
                     :func:`asyncio.get_event_loop` is used to get the default
                     event loop.
        """
        #: queue for consuming incoming event messages
        self.incoming_queue = incoming_queue
        #: CometD service url
        self._url = url
        #: http session
        self._http_session = http_session
        #: event loop used to schedule tasks
        self._loop = loop or asyncio.get_event_loop()
        #: clinet id value assigned by the server
        self._client_id = client_id
        #: message id which should be unique for every message during a client
        #: session
        self._message_id = 0
        #: reconnection advice parameters returned by the server
        self._reconnect_advice: JsonObject = reconnect_advice or dict()
        #: set of subscribed channels
        self._subscriptions: Set[str] = set()
        #: boolean to mark whether to resubscribe on connect
        self._subscribe_on_connect = False
        #: dictionary of TransportState and asyncio.Event pairs
        self._state_events = {_: asyncio.Event() for _ in TransportState}
        #: current state of the transport
        self._state = TransportState.DISCONNECTED
        #: asyncio connection task
        self._connect_task: Optional[asyncio.Future[JsonObject]] = None
        #: time to wait before reconnecting after a network failure
        self._reconnect_timeout = reconnection_timeout
        #: SSL validation mode
        self.ssl = ssl
        #: List of protocol extension objects
        self._extensions = extensions or []
        #: An auth extension
        self._auth = auth
        #: Function for JSON serialization
        self._json_dumps = json_dumps
        #: Function for JSON deserialization
        self._json_loads = json_loads

    @property
    def connection_type(self) -> ConnectionType:  # pragma: no cover
        """The transport's connection type"""
        return None  # type: ignore

    @property
    def endpoint(self) -> str:
        """CometD service url"""
        return self._url

    @property
    def client_id(self) -> Optional[str]:
        """clinet id value assigned by the server"""
        return self._client_id

    @property
    def subscriptions(self) -> Set[str]:
        """Set of subscribed channels"""
        return self._subscriptions

    @property
    def last_connect_result(self) -> Optional[JsonObject]:
        """Result of the last connect request"""
        if self._connect_task and self._connect_task.done():
            return self._connect_task.result()
        return None

    @property
    def reconnect_advice(self) -> JsonObject:
        """Reconnection advice parameters returned by the server"""
        return self._reconnect_advice

    @property
    def state(self) -> TransportState:
        """Current state of the transport"""
        return self._state

    @property
    def _state(self) -> TransportState:
        """Current state of the transport"""
        return self.__dict__.get("_state", TransportState.DISCONNECTED)

    @_state.setter
    def _state(self, value: TransportState) -> None:
        self._set_state_event(self._state, value)
        self.__dict__["_state"] = value

    @property
    def request_timeout(self) -> Optional[Union[int, float]]:
        """Number of seconds after a network request should time out"""
        timeout = self.reconnect_advice.get("timeout")
        if isinstance(timeout, (int, float)):
            # convert milliseconds to seconds
            timeout /= 1000
            # increase the timeout specified by the server to avoid timing out
            # by mistake
            timeout *= self.REQUEST_TIMEOUT_INCREASE_FACTOR
            return timeout
        return None

    def _set_state_event(self, old_state: TransportState,
                         new_state: TransportState) -> None:
        """Set event associated with the *new_state* and clear the event for
        the *old_state*

        :param old_state: Old state value
        :param new_state: New state value
        """
        if new_state != old_state:
            self._state_events[old_state].clear()
            self._state_events[new_state].set()

    async def wait_for_state(self, state: TransportState) -> None:
        """Waits for and returns when the transport enters the given *state*

        :param TransportState state: A state value
        """
        await self._state_events[state].wait()

    async def handshake(self, connection_types: List[ConnectionType]) \
            -> JsonObject:
        """Executes the handshake operation

        :param connection_types: list of connection types
        :return: Handshake response
        :raises TransportError: When the network request fails.
        """
        # reset message id for a new client session
        self._message_id = 0

        connection_types = list(connection_types)
        # make sure that the supported connection types list contains this
        # transport
        if self.connection_type not in connection_types:
            connection_types.append(self.connection_type)
        connection_type_values = [ct.value for ct in connection_types]

        # send message and await its response
        response_message = await self._send_message(
            HANDSHAKE_MESSAGE.copy(),
            supportedConnectionTypes=connection_type_values
        )
        # store the returned client id or set it to None if it's not in the
        # response
        if response_message["successful"]:
            self._client_id = response_message.get("clientId")
            self._subscribe_on_connect = True
        return response_message

    def _finalize_message(self, message: JsonObject) -> None:
        """Update the ``id``, ``clientId`` and ``connectionType`` message
        fields as a side effect if they're are present in the *message*.

        :param message: Outgoing message
        """
        if "id" in message:
            message["id"] = str(self._message_id)
            self._message_id += 1

        if "clientId" in message:
            message["clientId"] = self.client_id

        if "connectionType" in message:
            message["connectionType"] = self.connection_type.value

    def _finalize_payload(self, payload: Union[JsonObject, Payload]) \
            -> None:
        """Update the ``id``, ``clientId`` and ``connectionType`` message
        fields in the *payload*, as a side effect if they're are present in
        the *message*. The *payload* can be either a single message or a list
        of messages.

        :param payload: A message or a list of messages
        """
        if isinstance(payload, list):
            for item in payload:
                self._finalize_message(item)
        else:
            self._finalize_message(payload)

    async def _send_message(self, message: JsonObject, **kwargs: Any) \
            -> JsonObject:
        """Send message to server

        :param message: A message
        :param kwargs: Optional key-value pairs that'll be used to update the \
        the values in the *message*
        :return: Response message
        :raises TransportError: When the network request fails.
        """
        message.update(kwargs)
        return await self._send_payload_with_auth([message])

    async def _send_payload_with_auth(self, payload: Payload) \
            -> JsonObject:
        """Finalize and send *payload* to server and retry on authentication
        failure

        Finalize and send the *payload* to the server and return once a
        response message can be provided for the first message in the
        *payload*.

        :param payload: A list of messages
        :return: The response message for the first message in the *payload*
        :raises TransportError: When the network request fails.
        """
        response = await self._send_payload(payload)

        # if there is an auth extension and the response is an auth error
        if self._auth and is_auth_error_message(response):
            # then try to authenticate and resend the payload
            await self._auth.authenticate()
            return await self._send_payload(payload)

        # otherwise return the response
        return response

    async def _send_payload(self, payload: Payload) -> JsonObject:
        """Finalize and send *payload* to server

        Finalize and send the *payload* to the server and return once a
        response message can be provided for the first message in the
        *payload*.

        :param payload: A list of messages
        :return: The response message for the first message in the *payload*
        :raises TransportError: When the network request fails.
        """
        self._finalize_payload(payload)
        headers: Headers = {}
        # process the outgoing payload with the extensions
        await self._process_outgoing_payload(payload, headers)
        # send the payload to the server
        return await self._send_final_payload(payload, headers=headers)

    async def _process_outgoing_payload(self, payload: Payload,
                                        headers: Headers) -> None:
        """Process the outgoing *payload* and *headers* with the extensions

        :param payload: A list of messages
        :param headers: Headers to send
        """
        for extension in self._extensions:
            await extension.outgoing(payload, headers)
        if self._auth:
            await self._auth.outgoing(payload, headers)

    @abstractmethod
    async def _send_final_payload(self, payload: Payload, *,
                                  headers: Headers) -> JsonObject:
        """Send *payload* to server

        Send the *payload* to the server and return once a
        response message can be provided for the first message in the
        *payload*.
        When reimplementing this method keep in mind that the server will
        likely return additional responses which should be enqueued for
        consumers. To enqueue the received messages :meth:`_consume_payload`
        can be used.

        :param payload: A list of messages
        :param headers: Headers to send
        :return: The response message for the first message in the *payload*
        :raises TransportError: When the network request fails.
        """

    async def _consume_message(self, response_message: JsonObject) -> None:
        """Enqueue the *response_message* for consumers if it's a type of
        message that consumers should receive

        :param response_message: A response message
        """
        if is_event_message(response_message):
            await self.incoming_queue.put(response_message)

    def _update_subscriptions(self, response_message: JsonObject) -> None:
        """Update the set of subscriptions based on the *response_message*

       :param response_message: A response message
        """
        # if a subscription response is successful, then add the channel
        # to the set of subscriptions, if it fails, then remove it
        if response_message["channel"] == MetaChannel.SUBSCRIBE:
            if (response_message["successful"] and
                    response_message["subscription"]
                    not in self._subscriptions):
                self._subscriptions.add(response_message["subscription"])
            elif (not response_message["successful"] and
                  "subscription" in response_message and
                  response_message["subscription"] in self._subscriptions):
                self._subscriptions.remove(response_message["subscription"])

        # if an unsubscribe response is successful then remove the channel
        # from the set of subscriptions
        if response_message["channel"] == MetaChannel.UNSUBSCRIBE:
            if (response_message["successful"] and
                    response_message["subscription"] in self._subscriptions):
                self._subscriptions.remove(response_message["subscription"])

    async def _process_incoming_payload(self, payload: Payload,
                                        headers: Optional[Headers] = None) \
            -> None:
        """Process incoming *payload* and *headers* with the extensions

        :param payload: A list of response messages
        :param headers: Received headers
        """
        if self._auth:
            await self._auth.incoming(payload, headers)
        for extension in self._extensions:
            await extension.incoming(payload, headers)

    async def _consume_payload(
            self, payload: Payload, *,
            headers: Optional[Headers] = None,
            find_response_for: Optional[JsonObject] = None) \
            -> Optional[JsonObject]:
        """Enqueue event messages for the consumers and update the internal
        state of the transport, based on response messages in the *payload*.

        :param payload: A list of response messages
        :param headers: Received headers
        :param find_response_for: Find and return the matching \
        response message for the given *find_response_for* message.
        :return: The response message for the *find_response_for* message, \
        otherwise ``None``
        """
        # process incoming payload and headers with the extensions
        await self._process_incoming_payload(payload, headers)

        # return None if no response message is found for *find_response_for*
        result = None
        for message in payload:
            # if there is an advice in the message then update the transport's
            # reconnect advice
            if "advice" in message:
                self._reconnect_advice = message["advice"]

            # update subscriptions based on responses
            self._update_subscriptions(message)

            # set the message as the result and continue if it is a matching
            # response
            if (result is None and
                    is_matching_response(message, find_response_for)):
                result = message
                continue

            await self._consume_message(message)
        return result

    def _start_connect_task(self, coro: Awaitable[JsonObject]) \
            -> Awaitable[JsonObject]:
        """Wrap the *coro* in a future and schedule it

        The future is stored internally in :obj:`_connect_task`. The future's
        results will be consumed by :obj:`_connect_done`.

        :param coro: Coroutine
        :return: Future
        """
        self._connect_task = asyncio.ensure_future(coro, loop=self._loop)
        self._connect_task.add_done_callback(self._connect_done)
        return self._connect_task

    async def _stop_connect_task(self) -> None:
        """Stop the connection task

        If no connect task exists or if it's done it does nothing.
        """
        if self._connect_task and not self._connect_task.done():
            self._connect_task.cancel()
            await asyncio.wait([self._connect_task])

    async def connect(self) -> JsonObject:
        """Connect to the server

        The transport will try to start and maintain a continuous connection
        with the server, but it'll return with the response of the first
        successful connection as soon as possible.

        :return: The response of the first successful connection.
        :raise TransportInvalidOperation: If the transport doesn't has a \
        client id yet, or if it's not in a :obj:`~TransportState.DISCONNECTED`\
        :obj:`state`.
        :raises TransportError: When the network request fails.
        """
        if not self.client_id:
            raise TransportInvalidOperation(
                "Can't connect to the server without a client id. "
                "Do a handshake first.")
        if self.state not in [TransportState.DISCONNECTED,
                              TransportState.SERVER_DISCONNECTED]:
            raise TransportInvalidOperation(
                "Can't connect to a server without disconnecting first.")

        self._state = TransportState.CONNECTING
        return await self._start_connect_task(self._connect())

    async def _connect(self) -> JsonObject:
        """Connect to the server

        :return: Connect response
        :raises TransportError: When the network request fails.
        """
        message = CONNECT_MESSAGE.copy()
        payload = [message]
        if self._subscribe_on_connect and self.subscriptions:
            for subscription in self.subscriptions:
                extra_message = SUBSCRIBE_MESSAGE.copy()
                extra_message["subscription"] = subscription  # type: ignore
                payload.append(extra_message)
        result = await self._send_payload_with_auth(payload)
        self._subscribe_on_connect = not result["successful"]
        return result

    def _connect_done(self, future: "asyncio.Future[JsonObject]") -> None:
        """Consume the result of the *future* and follow the server's \
        connection advice if the transport is still connected

        :param future: A :obj:`_connect` or :obj:`handshake` \
        future
        """
        # set the default reconnect advice value
        reconnect_advice = "retry"
        # use the last known connection timeout
        reconnect_timeout = self.reconnect_advice.get("interval")
        try:
            # get the task's result
            result: Union[JsonObject, Exception] = future.result()
            # on a failed connect or handshake operation use the reconnect
            # advice returned by the server if it's present int the response
            # message
            if (isinstance(result, dict) and
                    not result.get("successful", True) and
                    "advice" in result and
                    "reconnect" in result["advice"]):
                reconnect_advice = result["advice"]["reconnect"]
            self._state = TransportState.CONNECTED
        except Exception as error:  # pylint: disable=broad-except
            result = error
            reconnect_timeout = self._reconnect_timeout
            if self.state != TransportState.DISCONNECTING:
                self._state = TransportState.CONNECTING

        LOGGER.debug("Connect task finished with: %r", result)

        if self.state != TransportState.DISCONNECTING:
            self._follow_advice(reconnect_advice, reconnect_timeout)

    def _follow_advice(self, reconnect_advice: str,
                       reconnect_timeout: Union[int, float, None]) -> None:
        """Follow the server's reconnect advice

        Either a :obj:`_connect` or :obj:`handshake` operation is started
        based on the *reconnect_advice* or the method returns without starting
        any operation if a different advice is specified.

        :param reconnect_advice: Reconnect advice parameter that determines \
        which operation should be started.
        :param reconnect_timeout: Initial connection delay to pass to \
        :obj:`_connect` or :obj:`handshake`.
        """
        # do a handshake operation if advised
        if reconnect_advice == "handshake":
            handshake_coro = defer(self.handshake,
                                   delay=reconnect_timeout,
                                   loop=self._loop)
            self._start_connect_task(handshake_coro([self.connection_type]))

        # do a connect operation if advised
        elif reconnect_advice == "retry":
            connect_coro = defer(self._connect,
                                 delay=reconnect_timeout,
                                 loop=self._loop)
            self._start_connect_task(connect_coro())

        # there is not reconnect advice from the server or its value
        # is none
        else:
            LOGGER.warning("No reconnect advice provided, no more operations "
                           "will be scheduled.")
            self._state = TransportState.SERVER_DISCONNECTED

    async def disconnect(self) -> None:
        """Disconnect from server

        The disconnect message is only sent to the server if the transport is
        actually connected.
        """
        try:
            should_send_message = self.state == TransportState.CONNECTED

            self._state = TransportState.DISCONNECTING
            await self._stop_connect_task()

            if should_send_message:
                with suppress(TransportError):
                    await self._send_message(DISCONNECT_MESSAGE.copy())
        finally:
            self._state = TransportState.DISCONNECTED

    async def close(self) -> None:
        """Close transport and release resources"""

    async def subscribe(self, channel: str) -> JsonObject:
        """Subscribe to *channel*

        :param channel: Name of the channel
        :return: Subscribe response
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """
        if self.state not in [TransportState.CONNECTING,
                              TransportState.CONNECTED]:
            raise TransportInvalidOperation(
                "Can't subscribe without being connected to a server.")
        return await self._send_message(SUBSCRIBE_MESSAGE.copy(),
                                        subscription=channel)

    async def unsubscribe(self, channel: str) -> JsonObject:
        """Unsubscribe from *channel*

        :param channel: Name of the channel
        :return: Unsubscribe response
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """
        if self.state not in [TransportState.CONNECTING,
                              TransportState.CONNECTED]:
            raise TransportInvalidOperation(
                "Can't unsubscribe without being connected to a server.")
        return await self._send_message(UNSUBSCRIBE_MESSAGE.copy(),
                                        subscription=channel)

    async def publish(self, channel: str, data: JsonObject) -> JsonObject:
        """Publish *data* to the given *channel*

        :param channel: Name of the channel
        :param data: Data to send to the server
        :return: Publish response
        :raise TransportInvalidOperation: If the transport is not in the \
        :obj:`~TransportState.CONNECTED` or :obj:`~TransportState.CONNECTING` \
        :obj:`state`
        :raises TransportError: When the network request fails.
        """
        if self.state not in [TransportState.CONNECTING,
                              TransportState.CONNECTED]:
            raise TransportInvalidOperation(
                "Can't publish without being connected to a server.")
        return await self._send_message(PUBLISH_MESSAGE.copy(),
                                        channel=channel,
                                        data=data)



"""Transport related constants"""
from enum import Enum, unique, auto


@unique
class ConnectionType(Enum):
    """CometD Connection types"""
    #: Long polling connection type
    LONG_POLLING = "long-polling"
    #: Websocket connection type
    WEBSOCKET = "websocket"


#: Connection type that all CometD server implementaions should support
DEFAULT_CONNECTION_TYPE = ConnectionType.LONG_POLLING
#: CometD meta channel prefix
META_CHANNEL_PREFIX = "/meta/"
#: CometD service channel prefix
SERVICE_CHANNEL_PREFIX = "/service/"


@unique
class MetaChannel(str, Enum):
    """CometD meta channel names"""
    #: Handshake meta channel
    HANDSHAKE = META_CHANNEL_PREFIX + "handshake"
    #: Connect meta channel
    CONNECT = META_CHANNEL_PREFIX + "connect"
    #: Disconnect meta channel
    DISCONNECT = META_CHANNEL_PREFIX + "disconnect"
    #: Subscribe meta channel
    SUBSCRIBE = META_CHANNEL_PREFIX + "subscribe"
    #: Unsubscribe meta channel
    UNSUBSCRIBE = META_CHANNEL_PREFIX + "unsubscribe"


@unique
class TransportState(Enum):
    """Describes a transport object's state"""
    #: Transport is disconnected
    DISCONNECTED = auto()
    #: Connection terminated by the server
    SERVER_DISCONNECTED = auto()
    #: Transport is trying to establish a connection
    CONNECTING = auto()
    #: Transport is connected to the server
    CONNECTED = auto()
    #: Transport is disconnecting from the server
    DISCONNECTING = auto()


#: Handshake message template
HANDSHAKE_MESSAGE = {
    # mandatory
    "channel": MetaChannel.HANDSHAKE,
    "version": "1.0",
    "supportedConnectionTypes": None,
    # optional
    "minimumVersion": "1.0",
    "id": None
}

#: Connect message template
CONNECT_MESSAGE = {
    # mandatory
    "channel": MetaChannel.CONNECT,
    "clientId": None,
    "connectionType": None,
    # optional
    "id": None
}

#: Disconnect message template
DISCONNECT_MESSAGE = {
    # mandatory
    "channel": MetaChannel.DISCONNECT,
    "clientId": None,
    # optional
    "id": None
}

#: Subscribe message template
SUBSCRIBE_MESSAGE = {
    # mandatory
    "channel": MetaChannel.SUBSCRIBE,
    "clientId": None,
    "subscription": None,
    # optional
    "id": None
}

#: Unsubscribe message template
UNSUBSCRIBE_MESSAGE = {
    # mandatory
    "channel": MetaChannel.UNSUBSCRIBE,
    "clientId": None,
    "subscription": None,
    # optional
    "id": None
}

#: Publish message template
PUBLISH_MESSAGE = {
    # mandatory
    "channel": None,
    "clientId": None,
    "data": None,
    # optional
    "id": None
}



"""Long polling transport class definition"""
import asyncio
import logging
from typing import Any

import aiohttp

from ..constants import ConnectionType
from ..exceptions import TransportError
from ..typing import JsonObject
from .registry import register_transport
from .base import TransportBase, Payload, Headers


LOGGER = logging.getLogger(__name__)


@register_transport(ConnectionType.LONG_POLLING)
class LongPollingTransport(TransportBase):
    """Long-polling type transport"""
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)

        #: semaphore to limit the number of concurrent HTTP connections to 2
        self._http_semaphore = asyncio.Semaphore(2)

    async def _send_final_payload(self, payload: Payload, *,
                                  headers: Headers) -> JsonObject:
        try:
            session = self._http_session
            async with self._http_semaphore:
                response = await session.post(self._url, json=payload,
                                              ssl=self.ssl, headers=headers,
                                              timeout=self.request_timeout)
            response_payload = await response.json(loads=self._json_loads)
            headers = response.headers
        except aiohttp.ClientError as error:
            LOGGER.warning("Failed to send payload, %s", error)
            raise TransportError(str(error)) from error
        response_message = await self._consume_payload(
            response_payload,
            headers=headers,
            find_response_for=payload[0]
        )

        if response_message is None:
            error_message = "No response message received for the " \
                            "first message in the payload"
            LOGGER.warning(error_message)
            raise TransportError(error_message)
        return response_message
